package jdk.net;

import java.security.AccessController;
import java.security.PrivilegedAction;
import java.net.MulticastSocket;
import java.util.Collections;
import java.net.StandardSocketOptions;
import java.util.HashSet;
import java.net.ServerSocket;
import java.net.Socket;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.DatagramSocket;
import java.lang.reflect.Method;
import java.net.SocketOption;
import java.util.Set;
import java.util.HashMap;
import jdk.Exported;

@Exported
public class Sockets
{
    private static final HashMap<Class<?>, Set<SocketOption<?>>> options;
    private static Method siSetOption;
    private static Method siGetOption;
    private static Method dsiSetOption;
    private static Method dsiGetOption;
    
    private static void initMethods() {
        try {
            Class<Object> forName = Class.forName("java.net.SocketSecrets");
            (Sockets.siSetOption = forName.getDeclaredMethod("setOption", new Class[] { Object.class, SocketOption.class, Object.class })).setAccessible(true);
            (Sockets.siGetOption = forName.getDeclaredMethod("getOption", new Class[] { Object.class, SocketOption.class })).setAccessible(true);
            (Sockets.dsiSetOption = forName.getDeclaredMethod("setOption", new Class[] { DatagramSocket.class, SocketOption.class, Object.class })).setAccessible(true);
            (Sockets.dsiGetOption = forName.getDeclaredMethod("getOption", new Class[] { DatagramSocket.class, SocketOption.class })).setAccessible(true);
        }
        catch (ReflectiveOperationException cause) {
            throw new InternalError(cause);
        }
    }
    
    private static <T> void invokeSet(Method method, Object o, SocketOption<T> socketOption, T t) throws IOException {
        try {
            method.invoke(null, new Object[] { o, socketOption, t });
        }
        catch (Exception cause) {
            if ((cause instanceof InvocationTargetException)) {
                Throwable targetException = ((InvocationTargetException)cause).getTargetException();
                if ((targetException instanceof IOException)) {
                    throw (IOException)targetException;
                }
                if ((targetException instanceof RuntimeException)) {
                    throw (RuntimeException)targetException;
                }
            }
            throw new RuntimeException(cause);
        }
    }
    
    private static <T> T invokeGet(Method method, Object o, SocketOption<T> socketOption) throws IOException {
        try {
            return method.invoke(null, new Object[] { o, socketOption });
        }
        catch (Exception cause) {
            if ((cause instanceof InvocationTargetException)) {
                Throwable targetException = ((InvocationTargetException)cause).getTargetException();
                if ((targetException instanceof IOException)) {
                    throw (IOException)targetException;
                }
                if ((targetException instanceof RuntimeException)) {
                    throw (RuntimeException)targetException;
                }
            }
            throw new RuntimeException(cause);
        }
    }
    
    private Sockets() {
    }
    
    public static <T> void setOption(Socket socket, SocketOption<T> socketOption, T t) throws IOException {
        if (!isSupported(Socket.class, socketOption)) {
            throw new UnsupportedOperationException(socketOption.name());
        }
        invokeSet(Sockets.siSetOption, socket, socketOption, t);
        return;
    }
    
    public static <T> T getOption(Socket socket, SocketOption<T> socketOption) throws IOException {
        if (!isSupported(Socket.class, socketOption)) {
            throw new UnsupportedOperationException(socketOption.name());
        }
        return invokeGet(Sockets.siGetOption, socket, socketOption);
    }
    
    public static <T> void setOption(ServerSocket serverSocket, SocketOption<T> socketOption, T t) throws IOException {
        if (!isSupported(ServerSocket.class, socketOption)) {
            throw new UnsupportedOperationException(socketOption.name());
        }
        invokeSet(Sockets.siSetOption, serverSocket, socketOption, t);
        return;
    }
    
    public static <T> T getOption(ServerSocket serverSocket, SocketOption<T> socketOption) throws IOException {
        if (!isSupported(ServerSocket.class, socketOption)) {
            throw new UnsupportedOperationException(socketOption.name());
        }
        return invokeGet(Sockets.siGetOption, serverSocket, socketOption);
    }
    
    public static <T> void setOption(DatagramSocket datagramSocket, SocketOption<T> socketOption, T t) throws IOException {
        if (!isSupported(datagramSocket.getClass(), socketOption)) {
            throw new UnsupportedOperationException(socketOption.name());
        }
        invokeSet(Sockets.dsiSetOption, datagramSocket, socketOption, t);
        return;
    }
    
    public static <T> T getOption(DatagramSocket datagramSocket, SocketOption<T> socketOption) throws IOException {
        if (!isSupported(datagramSocket.getClass(), socketOption)) {
            throw new UnsupportedOperationException(socketOption.name());
        }
        return invokeGet(Sockets.dsiGetOption, datagramSocket, socketOption);
    }
    
    public static Set<SocketOption<?>> supportedOptions(Class<?> key) {
        Set<Object> set = (Set)Sockets.options.get(key);
        if (set == null) {
            throw new IllegalArgumentException("unknown socket type");
        }
        return set;
    }
    
    private static boolean isSupported(Class<?> clazz, SocketOption<?> socketOption) {
        return supportedOptions(clazz).contains(socketOption);
    }
    
    private static void initOptionSets() {
        int flowSupported = ExtendedSocketOptions.PlatformSocketOptions.get().flowSupported();
        int keepAliveOptionsSupported = ExtendedSocketOptions.PlatformSocketOptions.get().keepAliveOptionsSupported();
        HashSet s = new HashSet();
        s.add(StandardSocketOptions.SO_KEEPALIVE);
        s.add(StandardSocketOptions.SO_SNDBUF);
        s.add(StandardSocketOptions.SO_RCVBUF);
        s.add(StandardSocketOptions.SO_REUSEADDR);
        s.add(StandardSocketOptions.SO_LINGER);
        s.add(StandardSocketOptions.IP_TOS);
        s.add(StandardSocketOptions.TCP_NODELAY);
        if (!(flowSupported == 0)) {
            s.add(ExtendedSocketOptions.SO_FLOW_SLA);
        }
        if (!(keepAliveOptionsSupported == 0)) {
            s.add(ExtendedSocketOptions.TCP_KEEPCOUNT);
            s.add(ExtendedSocketOptions.TCP_KEEPIDLE);
            s.add(ExtendedSocketOptions.TCP_KEEPINTERVAL);
        }
        Sockets.options.put(Socket.class, Collections.unmodifiableSet(s));
        HashSet s2 = new HashSet();
        s2.add(StandardSocketOptions.SO_RCVBUF);
        s2.add(StandardSocketOptions.SO_REUSEADDR);
        s2.add(StandardSocketOptions.IP_TOS);
        if (!(keepAliveOptionsSupported == 0)) {
            s2.add(ExtendedSocketOptions.TCP_KEEPCOUNT);
            s2.add(ExtendedSocketOptions.TCP_KEEPIDLE);
            s2.add(ExtendedSocketOptions.TCP_KEEPINTERVAL);
        }
        Sockets.options.put(ServerSocket.class, Collections.unmodifiableSet(s2));
        HashSet s3 = new HashSet();
        s3.add(StandardSocketOptions.SO_SNDBUF);
        s3.add(StandardSocketOptions.SO_RCVBUF);
        s3.add(StandardSocketOptions.SO_REUSEADDR);
        s3.add(StandardSocketOptions.IP_TOS);
        if (!(flowSupported == 0)) {
            s3.add(ExtendedSocketOptions.SO_FLOW_SLA);
        }
        Sockets.options.put(DatagramSocket.class, Collections.unmodifiableSet(s3));
        HashSet s4 = new HashSet();
        s4.add(StandardSocketOptions.SO_SNDBUF);
        s4.add(StandardSocketOptions.SO_RCVBUF);
        s4.add(StandardSocketOptions.SO_REUSEADDR);
        s4.add(StandardSocketOptions.IP_TOS);
        s4.add(StandardSocketOptions.IP_MULTICAST_IF);
        s4.add(StandardSocketOptions.IP_MULTICAST_TTL);
        s4.add(StandardSocketOptions.IP_MULTICAST_LOOP);
        if (!(flowSupported == 0)) {
            s4.add(ExtendedSocketOptions.SO_FLOW_SLA);
        }
        Sockets.options.put(MulticastSocket.class, Collections.unmodifiableSet(s4));
    }
    
    static void access$000() {
        initMethods();
    }
    
    static {
        options = new HashMap();
        initOptionSets();
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                Sockets.access$000();
                return null;
            }
        });
    }
}
